﻿using System;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading;
using System.Xml.Linq;
using System.Xml.XPath;
using OpenSSL.Core;
using OpenSSL.Crypto;
using OpenSSL.X509;
using Org.BouncyCastle.Crypto.Digests;
using Org.BouncyCastle.Crypto.Engines;
using Org.BouncyCastle.Crypto.Parameters;
using Org.BouncyCastle.Crypto.Signers;
using Org.BouncyCastle.Math;


namespace PoskytovatelCasovejPeciatky
{
    class Program
    {
        static void Main(string[] args)
        {
            if ((args.Length != 2) || (args[0] != "--spustit-server") || (args[0] != "-s"))
            {
                Console.WriteLine("Poskytovatel sluzby casovej peciatky");
                Console.WriteLine("Nezadany parameter\nDostupne parametre:\n--spustit-server (-s) <heslo sukromneho kluca na tvorbu casovej peciatky>");
                return;
            }

            Server TimeStampServer = new Server(args[1]);
        }
    }

     

  class Server
  {
    string configFile = "pcpConfig.xml";

    XDocument config;
    int onlineServerPort;
    IPAddress networkServerAddress;
    X509Certificate timestampCert;
    CryptoKey timestampCertPrivateKey;
    string timestampCertPrivateKeyPassword;

    private TcpListener tcpListener;
    private Thread listenThread;

    public static BigInteger opensslBigToBigInteger(String opensslInt)
    {
        try
        {
            String hexInt = opensslInt.Replace(":", "");
            hexInt = hexInt.Replace("\\s+", "");
            return new BigInteger(hexInt, 16);
        }
        catch
        {
            return null;
        }
    }  

    private byte[] timestampSign(byte[] subject, RSA rsa)
    {
        RsaEngine rsaEngine = new RsaEngine();
        try
        {
            RsaKeyParameters signKey = new RsaKeyParameters(true, opensslBigToBigInteger(rsa.PublicModulus.ToString()), opensslBigToBigInteger(rsa.PrivateExponent.ToString()));
            PssSigner voteSigner = new PssSigner(rsaEngine, new Sha512Digest());
            voteSigner.Init(true, signKey);
            voteSigner.BlockUpdate(subject, 0, subject.Length);
            return voteSigner.GenerateSignature();
        }
        catch
        {
            Console.WriteLine(DateTime.Now.ToString() + " Chyba: Chyba pri tvorbe casovej peciatky");
            return null;
        }
    }

    public Server(string timestampCertPrivateKeyPassword)
    {
        Console.WriteLine("Poskytovatel sluzby casovej peciatky");

        try
        {
            config = XDocument.Load(configFile);
            onlineServerPort = int.Parse(config.XPathSelectElement("/pcpConfig/onlineServerPort").Value.Trim());
            networkServerAddress = IPAddress.Parse(config.XPathSelectElement("/pcpConfig/tempOnlineStorageServerAddress").Value.Trim());
            timestampCert = X509Certificate.FromDER(BIO.File(config.XPathSelectElement("/pcpConfig/timestampCert").Value.Trim(), "r"));
        }
        catch
        {
            Console.WriteLine("Chyba: Poskodeny alebo neexistujuci konfiguracny subor pcpConfig.xml");
            return;
        }
        try
        {
            timestampCertPrivateKey = CryptoKey.FromPrivateKey(BIO.File(config.XPathSelectElement("/pcpConfig/timestampCertPrivateKey").Value.Trim(), "r"), timestampCertPrivateKeyPassword);
        }
        catch
        {
            Console.WriteLine("Chyba: Nespravne heslo alebo poskodeny konfiguracny subor pcpConfig.xml");
            return;
        }
        

        try
        {
            this.tcpListener = new TcpListener(networkServerAddress, onlineServerPort);
            this.listenThread = new Thread(new ThreadStart(ListenForClients));
            this.listenThread.Start();
        }
        catch
        {
            Console.WriteLine("Chyba: Server nie je mozne spustit");
        }
    }

    private void ListenForClients()
    {
        this.tcpListener.Start();
        
        while (true)
        {
            try
            {
                TcpClient client = this.tcpListener.AcceptTcpClient();
                Thread clientThread = new Thread(new ParameterizedThreadStart(HandleClientComm));
                clientThread.Start(client);
            }
            catch
            {
                Console.WriteLine(DateTime.Now.ToString() + " Chyba: Chyba spojenia so Sietovym serverom (SiS)");
            }
        }        
    }

    private byte[] receiveConnectionReceivePhase(NetworkStream clientStream)
    {
        byte[] receivedData = new byte[0];
        byte[] message = new byte[1024];
        byte[] concat;
        int bytesRead;

        while (true)
        {
            bytesRead = 0;

            try
            {
                //blocks until a client sends a message                
                bytesRead = clientStream.Read(message, 0, 1024);
                concat = new byte[receivedData.Length + bytesRead];
                Buffer.BlockCopy(receivedData, 0, concat, 0, receivedData.Length);
                Buffer.BlockCopy(message, 0, concat, receivedData.Length, bytesRead);
                receivedData = concat;
            }
            catch
            {
                //a socket error has occured
                Console.WriteLine(DateTime.Now.ToString() + " Chyba: Chyba spojenia so Sietovym serverom (SiS)");
                receivedData = null;
                break;
            }

            if (bytesRead == 0)
            {
                //the client has disconnected from the server
                break;
            }


            //message has successfully been received
        }
        return receivedData;
    }

    private void receiveConnectionResponsePhase(NetworkStream clientStream, byte[] data)
    {
        try
        {
            clientStream.Write(data, 0, data.Length);
        }
        catch
        {
            Console.WriteLine(DateTime.Now.ToString() + " Chyba: Chyba spojenia so Sietovym serverom (SiS)");
        }
    }

    private void HandleClientComm(object client)
    {        
        TcpClient tcpClient = (TcpClient)client;
        NetworkStream clientStream;
        byte[] timestampMessage;
        try
        {
            clientStream = tcpClient.GetStream();
            timestampMessage = receiveConnectionReceivePhase(clientStream);
        }
        catch
        {
            Console.WriteLine(DateTime.Now.ToString() + " Chyba: Chyba spojenia so Sietovym serverom (SiS)");
            return;
        }

        if ((timestampMessage != null) && (timestampMessage.Length > 0))
        {
            bool tryAgain = false;
            StreamReader maxSavedSerialNumReadStream = null;
            FileStream maxSavedSerialNumFile = null;
            while (!tryAgain)
            {
                tryAgain = false;
                try
                {
                    maxSavedSerialNumFile = File.Open("maxSavedSerialNum.toss", FileMode.Open, FileAccess.ReadWrite, FileShare.None);
                    maxSavedSerialNumReadStream = new StreamReader(maxSavedSerialNumFile);
                }
                catch
                {
                    tryAgain = true;
                }
            }
            long serialNumber = long.Parse(maxSavedSerialNumReadStream.ReadLine()) + 1;
            StreamWriter maxSavedSerialNumWriteStream;
            try
            {
                maxSavedSerialNumFile.SetLength(0);
                maxSavedSerialNumWriteStream = new StreamWriter(maxSavedSerialNumFile);
                maxSavedSerialNumWriteStream.Write(serialNumber);
            }
            catch
            {
                Console.WriteLine(DateTime.Now.ToString() + " Chyba: Chyba spojenia so Sietovym serverom (SiS)");
                return;
            }
            maxSavedSerialNumWriteStream.Close();

            byte[] date = Encoding.ASCII.GetBytes(DateTime.Now.ToUniversalTime().ToString());
            byte[] serialNumberByte = Encoding.ASCII.GetBytes(serialNumber.ToString());
            byte[] certByte = timestampCert.DER;
            byte[] signatureData = new byte[timestampMessage.Length + date.Length + certByte.Length + serialNumberByte.Length];
            Buffer.BlockCopy(serialNumberByte, 0, signatureData, 0, serialNumberByte.Length);
            Buffer.BlockCopy(timestampMessage, 0, signatureData, serialNumberByte.Length, timestampMessage.Length);
            Buffer.BlockCopy(date, 0, signatureData, serialNumberByte.Length + timestampMessage.Length, date.Length);
            Buffer.BlockCopy(certByte, 0, signatureData, serialNumberByte.Length + timestampMessage.Length + date.Length, certByte.Length);

            XDocument timestampResponse = new XDocument();
            XElement root = new XElement("timestamp",
                                new XElement("serialNumber",Encoding.ASCII.GetString(serialNumberByte)),
                                new XElement("dataHash", Convert.ToBase64String(timestampMessage)),
                                new XElement("dateTime", Encoding.ASCII.GetString(date)),
                                new XElement("certificate",Convert.ToBase64String(timestampCert.DER)),
                                new XElement("signature", Convert.ToBase64String(timestampSign(signatureData, timestampCertPrivateKey.GetRSA())))
                                );
            timestampResponse.AddFirst(root);
            string timestampResponseMessage = timestampResponse.ToString();            
            receiveConnectionResponsePhase(clientStream, Encoding.ASCII.GetBytes(timestampResponseMessage));                           
        }

        tcpClient.Close();
    }
  }
}
